-----------------------------------------------------------------------------
-- Pred:		Predicates
-- 
-- Part of `Typing Haskell in Haskell', version of September 1, 1999
-- Copyright (c) Mark P Jones and the Oregon Graduate Institute
-- of Science and Technology, 1999
-- 
-- This program is distributed as Free Software under the terms
-- in the file "License" that is included in the distribution
-- of this software, copies of which may be obtained from:
--             http://www.cse.ogi.edu/~mpj/thih/
-- 
-----------------------------------------------------------------------------

module Pred where
import List(union)
import Id
import Type
import Subst
import Unify
import PPrint

data Qual t = [Pred] :=> t
              deriving Eq

instance PPrint t => PPrint (Qual t) where
  pprint (ps  :=> t) = (pprint ps <+> text ":=>") $$ nest 2 (parPprint t)

data Pred   = IsIn Class [Type]
              deriving Eq

instance PPrint Pred where
  pprint (IsIn c ts) = text "isIn1" <+> text ('c' : name c) <+> pplist ts

instance Types t => Types (Qual t) where
  apply s (ps :=> t) = apply s ps :=> apply s t
  tv (ps :=> t)      = tv ps `union` tv t

instance Types Pred where
  apply s (IsIn c ts) = IsIn c (apply s ts)
  tv (IsIn c ts)      = tv ts

-----------------------------------------------------------------------------

data Class = Class { name  :: Id,
                     csig  :: [Tyvar],
                     super :: [Pred],
                     insts :: [Inst] }
type Inst  = Qual Pred

instance Eq Class where
  c == c'  =  name c == name c'

bySuper              :: Pred -> [Pred]
bySuper p@(IsIn c ts)
 = p : concat (map bySuper supers)
   where supers = apply s (super c)
         s      = zip (csig c) ts

byInst             :: Pred -> Inst -> Maybe [Pred]
byInst p (ps :=> h) = do u <- match h p
                         Just (map (apply u) ps)

instance Match Pred where
  match (IsIn c ts) (IsIn c' ts')
        | c == c'   = match ts ts'
        | otherwise = Nothing

reducePred :: Pred -> Maybe [Pred]
reducePred p@(IsIn c t) = foldr (|||) Nothing poss
 where poss = map (byInst p) (insts c)
       Nothing ||| y = y
       Just x  ||| y = Just x

-----------------------------------------------------------------------------

(||-)   :: [Pred] -> Pred -> Bool
ps ||- p = any (p `elem`) (map bySuper ps) ||
           case reducePred p of
             Nothing -> False
             Just qs -> all (ps ||-) qs

-----------------------------------------------------------------------------
