-----------------------------------------------------------------------------
-- StdPrel:	Encoding of (simplified) Haskell Standard Prelude
-----------------------------------------------------------------------------

module StdPrel where
import Kind
import Type
import Subst
import Pred
import Scheme
import Assump
import TIMonad

-----------------------------------------------------------------------------
-- This module contains definitions that do not appear in the
-- typeset version of the paper.

-- We introduce the following definition because it will allow a smoother
-- transition to multiple parameter classes later on ...

isIn1 c t    = IsIn c [t]

mkInst ks = inst ts 
 where ts   = zipWith (\v k -> TVar (Tyvar v k)) vars ks
       vars = [ [c] | c <-['a'..'z'] ] ++
              [ c : show n | n <-[0::Int ..], c<-['a'..'z'] ]

asig = [ Tyvar "a" Star ]
ta   = TVar (Tyvar "a" Star)
msig = [ Tyvar "m" (Kfun Star Star) ]

-----------------------------------------------------------------------------
-- Standard Types:
-- Definitions for Char, List, (,) and (->) primitive types are in "Type.lhs":

tup2Cfun
 = "(,)" :>: (Forall [Star, Star]
              ([] :=> 
               (TGen 0 `fn` TGen 1 `fn` TAp (TAp tTuple2 (TGen 0)) (TGen 1))))

tup3Cfun
 = "(,,)" :>: (Forall [Star, Star, Star]
               ([] :=> 
                (TGen 0 `fn` TGen 1 `fn` TGen 2 `fn` TAp (TAp (TAp tTuple3 (TGen 0)) (TGen 1)) (TGen 2))))

tup4Cfun
 = "(,,,)" :>: (Forall [Star, Star, Star, Star]
                ([] :=> 
                 (TGen 0 `fn` TGen 1 `fn` TGen 2 `fn` TGen 3 `fn` TAp (TAp (TAp (TAp tTuple4 (TGen 0)) (TGen 1)) (TGen 2)) (TGen 3))))

tup5Cfun
 = "(,,,,)" :>: (Forall [Star, Star, Star, Star, Star]
                 ([] :=> 
                  (TGen 0 `fn` TGen 1 `fn` TGen 2 `fn` TGen 3 `fn` TGen 4 `fn` TAp (TAp (TAp (TAp (TAp tTuple5 (TGen 0)) (TGen 1)) (TGen 2)) (TGen 3)) (TGen 4))))

tup6Cfun
 = "(,,,,,)" :>: (Forall [Star, Star, Star, Star, Star, Star]
                  ([] :=> 
                   (TGen 0 `fn` TGen 1 `fn` TGen 2 `fn` TGen 3 `fn` TGen 4 `fn` TGen 5 `fn` TAp (TAp (TAp (TAp (TAp (TAp tTuple6 (TGen 0)) (TGen 1)) (TGen 2)) (TGen 3)) (TGen 4)) (TGen 5))))

tup7Cfun
 = "(,,,,,,)" :>: (Forall [Star, Star, Star, Star, Star, Star, Star]
                   ([] :=> 
                    (TGen 0 `fn` TGen 1 `fn` TGen 2 `fn` TGen 3 `fn` TGen 4 `fn` TGen 5 `fn` TGen 6 `fn` TAp (TAp (TAp (TAp (TAp (TAp (TAp tTuple7 (TGen 0)) (TGen 1)) (TGen 2)) (TGen 3)) (TGen 4)) (TGen 5)) (TGen 6))))

unitCfun
 = "()" :>: (Forall []
              ([] :=> 
               tUnit))

nilCfun
 = "[]" :>: (Forall [Star]
              ([] :=> 
               (TAp tList (TGen 0))))
consCfun
 = ":" :>: (Forall [Star]
             ([] :=> 
              (TGen 0 `fn` TAp tList (TGen 0) `fn` TAp tList (TGen 0))))

tReadS a
 = TAp tList tChar `fn` TAp tList (TAp (TAp tTuple2 a) (TAp tList tChar))

tShowS
 = TAp tList tChar `fn` TAp tList tChar

tBool
 = TCon (Tycon "Bool" Star)
falseCfun
 = "False" :>: (Forall []
                 ([] :=> 
                  tBool))
trueCfun
 = "True" :>: (Forall []
                ([] :=> 
                 tBool))

tString
 = TAp tList tChar

tMaybe
 = TCon (Tycon "Maybe" (Kfun Star Star))
nothingCfun
 = "Nothing" :>: (Forall [Star]
                   ([] :=> 
                    (TAp tMaybe (TGen 0))))
justCfun
 = "Just" :>: (Forall [Star]
                ([] :=> 
                 (TGen 0 `fn` TAp tMaybe (TGen 0))))

tEither
 = TCon (Tycon "Either" (Kfun Star (Kfun Star Star)))
leftCfun
 = "Left" :>: (Forall [Star, Star]
                ([] :=> 
                 (TGen 0 `fn` TAp (TAp tEither (TGen 0)) (TGen 1))))
rightCfun
 = "Right" :>: (Forall [Star, Star]
                 ([] :=> 
                  (TGen 1 `fn` TAp (TAp tEither (TGen 0)) (TGen 1))))

tOrdering
 = TCon (Tycon "Ordering" Star)
lTCfun
 = "LT" :>: (Forall []
              ([] :=> 
               tOrdering))
eQCfun
 = "EQ" :>: (Forall []
              ([] :=> 
               tOrdering))
gTCfun
 = "GT" :>: (Forall []
              ([] :=> 
               tOrdering))

tInt
 = TCon (Tycon "Int" Star)

tInteger
 = TCon (Tycon "Integer" Star)

tFloat
 = TCon (Tycon "Float" Star)

tDouble
 = TCon (Tycon "Double" Star)

tRatio
 = TCon (Tycon "Ratio" (Kfun Star Star))
consratCfun
 = ":%" :>: (Forall [Star]
              ([isIn1 cIntegral (TGen 0)] :=> 
               (TGen 0 `fn` TGen 0 `fn` TAp tRatio (TGen 0))))

tRational
 = TAp tRatio tInteger

tIOError
 = TCon (Tycon "IOError" Star)

tFilePath
 = TAp tList tChar

tIO
 = TCon (Tycon "IO" (Kfun Star Star))
iOCfun
 = "IO" :>: (Forall [Star]
              ([] :=> 
               (((tIOError `fn` TAp tIOResult (TGen 0)) `fn` (TGen 0 `fn` TAp tIOResult (TGen 0)) `fn` TAp tIOResult (TGen 0)) `fn` TAp tIO (TGen 0))))

tIOResult
 = TCon (Tycon "IOResult" (Kfun Star Star))
hugs_ExitWithCfun
 = "Hugs_ExitWith" :>: (Forall [Star]
                         ([] :=> 
                          (tInt `fn` TAp tIOResult (TGen 0))))
hugs_SuspendThreadCfun
 = "Hugs_SuspendThread" :>: (Forall [Star]
                              ([] :=> 
                               (TAp tIOResult (TGen 0))))
hugs_ErrorCfun
 = "Hugs_Error" :>: (Forall [Star]
                      ([] :=> 
                       (tIOError `fn` TAp tIOResult (TGen 0))))
hugs_ReturnCfun
 = "Hugs_Return" :>: (Forall [Star]
                       ([] :=> 
                        (TGen 0 `fn` TAp tIOResult (TGen 0))))

tId
 = TAp tList tChar

tKind
 = TCon (Tycon "Kind" Star)
starCfun
 = "Star" :>: (Forall []
                ([] :=> 
                 tKind))
kfunCfun
 = "Kfun" :>: (Forall []
                ([] :=> 
                 (tKind `fn` tKind `fn` tKind)))

tType
 = TCon (Tycon "Type" Star)
tVarCfun
 = "TVar" :>: (Forall []
                ([] :=> 
                 (tTyvar `fn` tType)))
tConCfun
 = "TCon" :>: (Forall []
                ([] :=> 
                 (tTycon `fn` tType)))
tApCfun
 = "TAp" :>: (Forall []
               ([] :=> 
                (tType `fn` tType `fn` tType)))
tGenCfun
 = "TGen" :>: (Forall []
                ([] :=> 
                 (tInt `fn` tType)))

tTyvar
 = TCon (Tycon "Tyvar" Star)
tyvarCfun
 = "Tyvar" :>: (Forall []
                 ([] :=> 
                  (tId `fn` tKind `fn` tTyvar)))

tTycon
 = TCon (Tycon "Tycon" Star)
tyconCfun
 = "Tycon" :>: (Forall []
                 ([] :=> 
                  (tId `fn` tKind `fn` tTycon)))

tSubst
 = TAp tList (TAp (TAp tTuple2 tTyvar) tType)

tQual
 = TCon (Tycon "Qual" (Kfun Star Star))
qualifyCfun
 = ":=>" :>: (Forall [Star]
               ([] :=> 
                (TAp tList tPred `fn` TGen 0 `fn` TAp tQual (TGen 0))))

tPred
 = TCon (Tycon "Pred" Star)
isInCfun
 = "IsIn" :>: (Forall []
                ([] :=> 
                 (tClass `fn` tType `fn` tPred)))

tClass
 = TCon (Tycon "Class" Star)
classCfun
 = "Class" :>: (Forall []
                 ([] :=> 
                  (tId `fn` TAp tList tClass `fn` TAp tList tInst `fn` tClass)))
nameSfun
 = "name" :>: (Forall []
                 ([] :=> 
                  (tClass `fn` tId)))
superSfun
 = "super" :>: (Forall []
                 ([] :=> 
                  (tClass `fn` TAp tList tClass)))
instsSfun
 = "insts" :>: (Forall []
                 ([] :=> 
                  (tClass `fn` TAp tList tInst)))

tInst
 = TAp tQual tPred

tScheme
 = TCon (Tycon "Scheme" Star)
forallCfun
 = "Forall" :>: (Forall []
                  ([] :=> 
                   (TAp tList tKind `fn` TAp tQual tType `fn` tScheme)))

tAssump
 = TCon (Tycon "Assump" Star)
assumeCfun
 = ":>:" :>: (Forall []
               ([] :=> 
                (tId `fn` tScheme `fn` tAssump)))

tTI
 = TCon (Tycon "TI" (Kfun Star Star))
tICfun
 = "TI" :>: (Forall [Star]
              ([] :=> 
               ((tSubst `fn` tInt `fn` TAp (TAp (TAp tTuple3 tSubst) tInt) (TGen 0)) `fn` TAp tTI (TGen 0))))

tInfer a b
 = TAp tList tAssump `fn` a `fn` TAp tTI (TAp (TAp tTuple2 (TAp tList tPred)) b)

tLit
 = TCon (Tycon "Lit" Star)
litIntCfun
 = "LitInt" :>: (Forall []
                  ([] :=> 
                   (tInteger `fn` tLit)))
litCharCfun
 = "LitChar" :>: (Forall []
                   ([] :=> 
                    (tChar `fn` tLit)))

tPat
 = TCon (Tycon "Pat" Star)
pVarCfun
 = "PVar" :>: (Forall []
                ([] :=> 
                 (tId `fn` tPat)))
pLitCfun
 = "PLit" :>: (Forall []
                ([] :=> 
                 (tLit `fn` tPat)))
pConCfun
 = "PCon" :>: (Forall []
                ([] :=> 
                 (tAssump `fn` TAp tList tPat `fn` tPat)))

tExpr
 = TCon (Tycon "Expr" Star)
varCfun
 = "Var" :>: (Forall []
               ([] :=> 
                (tId `fn` tExpr)))
litCfun
 = "Lit" :>: (Forall []
               ([] :=> 
                (tLit `fn` tExpr)))
constCfun
 = "Const" :>: (Forall []
                 ([] :=> 
                  (tAssump `fn` tExpr)))
apCfun
 = "Ap" :>: (Forall []
              ([] :=> 
               (tExpr `fn` tExpr `fn` tExpr)))
letCfun
 = "Let" :>: (Forall []
               ([] :=> 
                (tBindGroup `fn` tExpr `fn` tExpr)))

tAlt
 = TAp (TAp tTuple2 (TAp tList tPat)) tExpr

tExpl
 = TAp (TAp (TAp tTuple3 (TAp tList tChar)) tScheme) (TAp tList (TAp (TAp tTuple2 (TAp tList tPat)) tExpr))

tImpl
 = TAp (TAp tTuple2 (TAp tList tChar)) (TAp tList (TAp (TAp tTuple2 (TAp tList tPat)) tExpr))

tBindGroup
 = TAp (TAp tTuple2 (TAp tList (TAp (TAp (TAp tTuple3 (TAp tList tChar)) tScheme) (TAp tList (TAp (TAp tTuple2 (TAp tList tPat)) tExpr))))) (TAp tList (TAp tList (TAp (TAp tTuple2 (TAp tList tChar)) (TAp tList (TAp (TAp tTuple2 (TAp tList tPat)) tExpr)))))

tProgram
 = TAp tList (TAp (TAp tTuple2 (TAp tList (TAp (TAp (TAp tTuple3 (TAp tList tChar)) tScheme) (TAp tList (TAp (TAp tTuple2 (TAp tList tPat)) tExpr))))) (TAp tList (TAp tList (TAp (TAp tTuple2 (TAp tList tChar)) (TAp tList (TAp (TAp tTuple2 (TAp tList tPat)) tExpr))))))

cEq = Class { name = "Eq",
              csig = asig,
              super = [],
              insts = instsEq }

eqMfun
 = "==" :>: (Forall [Star]
              ([isIn1 cEq (TGen 0)] :=> 
               (TGen 0 `fn` TGen 0 `fn` tBool)))
neqMfun
 = "/=" :>: (Forall [Star]
              ([isIn1 cEq (TGen 0)] :=> 
               (TGen 0 `fn` TGen 0 `fn` tBool)))

instsEq
 = [mkInst []
     ([] :=> 
      isIn1 cEq tUnit),
    mkInst []
     ([] :=> 
      isIn1 cEq tChar),
    mkInst [Star]
     ([isIn1 cEq (TGen 0)] :=> 
      isIn1 cEq (TAp tList (TGen 0))),
    mkInst []
     ([] :=> 
      isIn1 cEq tInt),
    mkInst []
     ([] :=> 
      isIn1 cEq tInteger),
    mkInst []
     ([] :=> 
      isIn1 cEq tFloat),
    mkInst []
     ([] :=> 
      isIn1 cEq tDouble),
    mkInst []
     ([] :=> 
      isIn1 cEq tBool),
    mkInst [Star]
     ([isIn1 cEq (TGen 0)] :=> 
      isIn1 cEq (TAp tMaybe (TGen 0))),
    mkInst [Star, Star]
     ([isIn1 cEq (TGen 0),
       isIn1 cEq (TGen 1)] :=> 
      isIn1 cEq (TAp (TAp tEither (TGen 0)) (TGen 1))),
    mkInst []
     ([] :=> 
      isIn1 cEq tOrdering),
    mkInst [Star]
     ([isIn1 cIntegral (TGen 0)] :=> 
      isIn1 cEq (TAp tRatio (TGen 0))),
    mkInst [Star, Star, Star, Star, Star]
     ([isIn1 cEq (TGen 0),
       isIn1 cEq (TGen 1),
       isIn1 cEq (TGen 2),
       isIn1 cEq (TGen 3),
       isIn1 cEq (TGen 4)] :=> 
      isIn1 cEq (TAp (TAp (TAp (TAp (TAp tTuple5 (TGen 0)) (TGen 1)) (TGen 2)) (TGen 3)) (TGen 4))),
    mkInst [Star, Star, Star, Star]
     ([isIn1 cEq (TGen 0),
       isIn1 cEq (TGen 1),
       isIn1 cEq (TGen 2),
       isIn1 cEq (TGen 3)] :=> 
      isIn1 cEq (TAp (TAp (TAp (TAp tTuple4 (TGen 0)) (TGen 1)) (TGen 2)) (TGen 3))),
    mkInst [Star, Star, Star]
     ([isIn1 cEq (TGen 0),
       isIn1 cEq (TGen 1),
       isIn1 cEq (TGen 2)] :=> 
      isIn1 cEq (TAp (TAp (TAp tTuple3 (TGen 0)) (TGen 1)) (TGen 2))),
    mkInst [Star, Star]
     ([isIn1 cEq (TGen 0),
       isIn1 cEq (TGen 1)] :=> 
      isIn1 cEq (TAp (TAp tTuple2 (TGen 0)) (TGen 1))),
    mkInst []
     ([] :=> 
      isIn1 cEq tClass),
    mkInst []
     ([] :=> 
      isIn1 cEq tKind),
    mkInst []
     ([] :=> 
      isIn1 cEq tType),
    mkInst []
     ([] :=> 
      isIn1 cEq tTyvar),
    mkInst []
     ([] :=> 
      isIn1 cEq tTycon),
    mkInst [Star]
     ([isIn1 cEq (TGen 0)] :=> 
      isIn1 cEq (TAp tQual (TGen 0))),
    mkInst []
     ([] :=> 
      isIn1 cEq tPred),
    mkInst []
     ([] :=> 
      isIn1 cEq tScheme)]

cOrd = Class { name = "Ord",
               csig = asig,
               super = [isIn1 cEq ta],
               insts = instsOrd }

compareMfun
 = "compare" :>: (Forall [Star]
                   ([isIn1 cOrd (TGen 0)] :=> 
                    (TGen 0 `fn` TGen 0 `fn` tOrdering)))
ltMfun
 = "<" :>: (Forall [Star]
             ([isIn1 cOrd (TGen 0)] :=> 
              (TGen 0 `fn` TGen 0 `fn` tBool)))
leMfun
 = "<=" :>: (Forall [Star]
              ([isIn1 cOrd (TGen 0)] :=> 
               (TGen 0 `fn` TGen 0 `fn` tBool)))
geMfun
 = ">=" :>: (Forall [Star]
              ([isIn1 cOrd (TGen 0)] :=> 
               (TGen 0 `fn` TGen 0 `fn` tBool)))
gtMfun
 = ">" :>: (Forall [Star]
             ([isIn1 cOrd (TGen 0)] :=> 
              (TGen 0 `fn` TGen 0 `fn` tBool)))
maxMfun
 = "max" :>: (Forall [Star]
               ([isIn1 cOrd (TGen 0)] :=> 
                (TGen 0 `fn` TGen 0 `fn` TGen 0)))
minMfun
 = "min" :>: (Forall [Star]
               ([isIn1 cOrd (TGen 0)] :=> 
                (TGen 0 `fn` TGen 0 `fn` TGen 0)))

instsOrd
 = [mkInst []
     ([] :=> 
      isIn1 cOrd tUnit),
    mkInst []
     ([] :=> 
      isIn1 cOrd tChar),
    mkInst [Star]
     ([isIn1 cOrd (TGen 0)] :=> 
      isIn1 cOrd (TAp tList (TGen 0))),
    mkInst []
     ([] :=> 
      isIn1 cOrd tInt),
    mkInst []
     ([] :=> 
      isIn1 cOrd tInteger),
    mkInst []
     ([] :=> 
      isIn1 cOrd tFloat),
    mkInst []
     ([] :=> 
      isIn1 cOrd tDouble),
    mkInst [Star]
     ([isIn1 cIntegral (TGen 0)] :=> 
      isIn1 cOrd (TAp tRatio (TGen 0))),
    mkInst []
     ([] :=> 
      isIn1 cOrd tBool),
    mkInst [Star]
     ([isIn1 cOrd (TGen 0)] :=> 
      isIn1 cOrd (TAp tMaybe (TGen 0))),
    mkInst [Star, Star]
     ([isIn1 cOrd (TGen 0),
       isIn1 cOrd (TGen 1)] :=> 
      isIn1 cOrd (TAp (TAp tEither (TGen 0)) (TGen 1))),
    mkInst []
     ([] :=> 
      isIn1 cOrd tOrdering),
    mkInst [Star, Star, Star, Star, Star]
     ([isIn1 cOrd (TGen 0),
       isIn1 cOrd (TGen 1),
       isIn1 cOrd (TGen 2),
       isIn1 cOrd (TGen 3),
       isIn1 cOrd (TGen 4)] :=> 
      isIn1 cOrd (TAp (TAp (TAp (TAp (TAp tTuple5 (TGen 0)) (TGen 1)) (TGen 2)) (TGen 3)) (TGen 4))),
    mkInst [Star, Star, Star, Star]
     ([isIn1 cOrd (TGen 0),
       isIn1 cOrd (TGen 1),
       isIn1 cOrd (TGen 2),
       isIn1 cOrd (TGen 3)] :=> 
      isIn1 cOrd (TAp (TAp (TAp (TAp tTuple4 (TGen 0)) (TGen 1)) (TGen 2)) (TGen 3))),
    mkInst [Star, Star, Star]
     ([isIn1 cOrd (TGen 0),
       isIn1 cOrd (TGen 1),
       isIn1 cOrd (TGen 2)] :=> 
      isIn1 cOrd (TAp (TAp (TAp tTuple3 (TGen 0)) (TGen 1)) (TGen 2))),
    mkInst [Star, Star]
     ([isIn1 cOrd (TGen 0),
       isIn1 cOrd (TGen 1)] :=> 
      isIn1 cOrd (TAp (TAp tTuple2 (TGen 0)) (TGen 1)))]

cBounded = Class { name = "Bounded",
                   csig = asig,
                   super = [],
                   insts = instsBounded }

minBoundMfun
 = "minBound" :>: (Forall [Star]
                    ([isIn1 cBounded (TGen 0)] :=> 
                     (TGen 0)))
maxBoundMfun
 = "maxBound" :>: (Forall [Star]
                    ([isIn1 cBounded (TGen 0)] :=> 
                     (TGen 0)))

instsBounded
 = [mkInst []
     ([] :=> 
      isIn1 cBounded tUnit),
    mkInst []
     ([] :=> 
      isIn1 cBounded tChar),
    mkInst []
     ([] :=> 
      isIn1 cBounded tInt),
    mkInst []
     ([] :=> 
      isIn1 cBounded tBool),
    mkInst []
     ([] :=> 
      isIn1 cBounded tOrdering)]

cNum = Class { name = "Num",
               csig = asig,
               super = [isIn1 cEq ta,
                        isIn1 cShow ta],
               insts = instsNum }

plusMfun
 = "+" :>: (Forall [Star]
             ([isIn1 cNum (TGen 0)] :=> 
              (TGen 0 `fn` TGen 0 `fn` TGen 0)))
minusMfun
 = "-" :>: (Forall [Star]
             ([isIn1 cNum (TGen 0)] :=> 
              (TGen 0 `fn` TGen 0 `fn` TGen 0)))
timesMfun
 = "*" :>: (Forall [Star]
             ([isIn1 cNum (TGen 0)] :=> 
              (TGen 0 `fn` TGen 0 `fn` TGen 0)))
negateMfun
 = "negate" :>: (Forall [Star]
                  ([isIn1 cNum (TGen 0)] :=> 
                   (TGen 0 `fn` TGen 0)))
absMfun
 = "abs" :>: (Forall [Star]
               ([isIn1 cNum (TGen 0)] :=> 
                (TGen 0 `fn` TGen 0)))
signumMfun
 = "signum" :>: (Forall [Star]
                  ([isIn1 cNum (TGen 0)] :=> 
                   (TGen 0 `fn` TGen 0)))
fromIntegerMfun
 = "fromInteger" :>: (Forall [Star]
                       ([isIn1 cNum (TGen 0)] :=> 
                        (tInteger `fn` TGen 0)))
fromIntMfun
 = "fromInt" :>: (Forall [Star]
                   ([isIn1 cNum (TGen 0)] :=> 
                    (tInt `fn` TGen 0)))

instsNum
 = [mkInst []
     ([] :=> 
      isIn1 cNum tInt),
    mkInst []
     ([] :=> 
      isIn1 cNum tInteger),
    mkInst []
     ([] :=> 
      isIn1 cNum tFloat),
    mkInst []
     ([] :=> 
      isIn1 cNum tDouble),
    mkInst [Star]
     ([isIn1 cIntegral (TGen 0)] :=> 
      isIn1 cNum (TAp tRatio (TGen 0)))]

cReal = Class { name = "Real",
                csig = asig,
                super = [isIn1 cNum ta,
                         isIn1 cOrd ta],
                insts = instsReal }

toRationalMfun
 = "toRational" :>: (Forall [Star]
                      ([isIn1 cReal (TGen 0)] :=> 
                       (TGen 0 `fn` tRational)))

instsReal
 = [mkInst []
     ([] :=> 
      isIn1 cReal tInt),
    mkInst []
     ([] :=> 
      isIn1 cReal tInteger),
    mkInst []
     ([] :=> 
      isIn1 cReal tFloat),
    mkInst []
     ([] :=> 
      isIn1 cReal tDouble),
    mkInst [Star]
     ([isIn1 cIntegral (TGen 0)] :=> 
      isIn1 cReal (TAp tRatio (TGen 0)))]

cIntegral = Class { name = "Integral",
                    csig = asig,
                    super = [isIn1 cReal ta,
                             isIn1 cEnum ta],
                    insts = instsIntegral }

quotMfun
 = "quot" :>: (Forall [Star]
                ([isIn1 cIntegral (TGen 0)] :=> 
                 (TGen 0 `fn` TGen 0 `fn` TGen 0)))
remMfun
 = "rem" :>: (Forall [Star]
               ([isIn1 cIntegral (TGen 0)] :=> 
                (TGen 0 `fn` TGen 0 `fn` TGen 0)))
divMfun
 = "div" :>: (Forall [Star]
               ([isIn1 cIntegral (TGen 0)] :=> 
                (TGen 0 `fn` TGen 0 `fn` TGen 0)))
modMfun
 = "mod" :>: (Forall [Star]
               ([isIn1 cIntegral (TGen 0)] :=> 
                (TGen 0 `fn` TGen 0 `fn` TGen 0)))
quotRemMfun
 = "quotRem" :>: (Forall [Star]
                   ([isIn1 cIntegral (TGen 0)] :=> 
                    (TGen 0 `fn` TGen 0 `fn` TAp (TAp tTuple2 (TGen 0)) (TGen 0))))
divModMfun
 = "divMod" :>: (Forall [Star]
                  ([isIn1 cIntegral (TGen 0)] :=> 
                   (TGen 0 `fn` TGen 0 `fn` TAp (TAp tTuple2 (TGen 0)) (TGen 0))))
evenMfun
 = "even" :>: (Forall [Star]
                ([isIn1 cIntegral (TGen 0)] :=> 
                 (TGen 0 `fn` tBool)))
oddMfun
 = "odd" :>: (Forall [Star]
               ([isIn1 cIntegral (TGen 0)] :=> 
                (TGen 0 `fn` tBool)))
toIntegerMfun
 = "toInteger" :>: (Forall [Star]
                     ([isIn1 cIntegral (TGen 0)] :=> 
                      (TGen 0 `fn` tInteger)))
toIntMfun
 = "toInt" :>: (Forall [Star]
                 ([isIn1 cIntegral (TGen 0)] :=> 
                  (TGen 0 `fn` tInt)))

instsIntegral
 = [mkInst []
     ([] :=> 
      isIn1 cIntegral tInt),
    mkInst []
     ([] :=> 
      isIn1 cIntegral tInteger)]

cFractional = Class { name = "Fractional",
                      csig = asig,
                      super = [isIn1 cNum ta],
                      insts = instsFractional }

divideMfun
 = "/" :>: (Forall [Star]
             ([isIn1 cFractional (TGen 0)] :=> 
              (TGen 0 `fn` TGen 0 `fn` TGen 0)))
recipMfun
 = "recip" :>: (Forall [Star]
                 ([isIn1 cFractional (TGen 0)] :=> 
                  (TGen 0 `fn` TGen 0)))
fromRationalMfun
 = "fromRational" :>: (Forall [Star]
                        ([isIn1 cFractional (TGen 0)] :=> 
                         (tRational `fn` TGen 0)))
fromDoubleMfun
 = "fromDouble" :>: (Forall [Star]
                      ([isIn1 cFractional (TGen 0)] :=> 
                       (tDouble `fn` TGen 0)))

instsFractional
 = [mkInst []
     ([] :=> 
      isIn1 cFractional tFloat),
    mkInst []
     ([] :=> 
      isIn1 cFractional tDouble),
    mkInst [Star]
     ([isIn1 cIntegral (TGen 0)] :=> 
      isIn1 cFractional (TAp tRatio (TGen 0)))]

cFloating = Class { name = "Floating",
                    csig = asig,
                    super = [isIn1 cFractional ta],
                    insts = instsFloating }

piMfun
 = "pi" :>: (Forall [Star]
              ([isIn1 cFloating (TGen 0)] :=> 
               (TGen 0)))
expMfun
 = "exp" :>: (Forall [Star]
               ([isIn1 cFloating (TGen 0)] :=> 
                (TGen 0 `fn` TGen 0)))
logMfun
 = "log" :>: (Forall [Star]
               ([isIn1 cFloating (TGen 0)] :=> 
                (TGen 0 `fn` TGen 0)))
sqrtMfun
 = "sqrt" :>: (Forall [Star]
                ([isIn1 cFloating (TGen 0)] :=> 
                 (TGen 0 `fn` TGen 0)))
starstarMfun
 = "**" :>: (Forall [Star]
              ([isIn1 cFloating (TGen 0)] :=> 
               (TGen 0 `fn` TGen 0 `fn` TGen 0)))
logBaseMfun
 = "logBase" :>: (Forall [Star]
                   ([isIn1 cFloating (TGen 0)] :=> 
                    (TGen 0 `fn` TGen 0 `fn` TGen 0)))
sinMfun
 = "sin" :>: (Forall [Star]
               ([isIn1 cFloating (TGen 0)] :=> 
                (TGen 0 `fn` TGen 0)))
cosMfun
 = "cos" :>: (Forall [Star]
               ([isIn1 cFloating (TGen 0)] :=> 
                (TGen 0 `fn` TGen 0)))
tanMfun
 = "tan" :>: (Forall [Star]
               ([isIn1 cFloating (TGen 0)] :=> 
                (TGen 0 `fn` TGen 0)))
asinMfun
 = "asin" :>: (Forall [Star]
                ([isIn1 cFloating (TGen 0)] :=> 
                 (TGen 0 `fn` TGen 0)))
acosMfun
 = "acos" :>: (Forall [Star]
                ([isIn1 cFloating (TGen 0)] :=> 
                 (TGen 0 `fn` TGen 0)))
atanMfun
 = "atan" :>: (Forall [Star]
                ([isIn1 cFloating (TGen 0)] :=> 
                 (TGen 0 `fn` TGen 0)))
sinhMfun
 = "sinh" :>: (Forall [Star]
                ([isIn1 cFloating (TGen 0)] :=> 
                 (TGen 0 `fn` TGen 0)))
coshMfun
 = "cosh" :>: (Forall [Star]
                ([isIn1 cFloating (TGen 0)] :=> 
                 (TGen 0 `fn` TGen 0)))
tanhMfun
 = "tanh" :>: (Forall [Star]
                ([isIn1 cFloating (TGen 0)] :=> 
                 (TGen 0 `fn` TGen 0)))
asinhMfun
 = "asinh" :>: (Forall [Star]
                 ([isIn1 cFloating (TGen 0)] :=> 
                  (TGen 0 `fn` TGen 0)))
acoshMfun
 = "acosh" :>: (Forall [Star]
                 ([isIn1 cFloating (TGen 0)] :=> 
                  (TGen 0 `fn` TGen 0)))
atanhMfun
 = "atanh" :>: (Forall [Star]
                 ([isIn1 cFloating (TGen 0)] :=> 
                  (TGen 0 `fn` TGen 0)))

instsFloating
 = [mkInst []
     ([] :=> 
      isIn1 cFloating tFloat),
    mkInst []
     ([] :=> 
      isIn1 cFloating tDouble)]

cRealFrac = Class { name = "RealFrac",
                    csig = asig,
                    super = [isIn1 cReal ta,
                             isIn1 cFractional ta],
                    insts = instsRealFrac }

properFractionMfun
 = "properFraction" :>: (Forall [Star, Star]
                          ([isIn1 cRealFrac (TGen 0),
                            isIn1 cIntegral (TGen 1)] :=> 
                           (TGen 0 `fn` TAp (TAp tTuple2 (TGen 1)) (TGen 0))))
truncateMfun
 = "truncate" :>: (Forall [Star, Star]
                    ([isIn1 cRealFrac (TGen 0),
                      isIn1 cIntegral (TGen 1)] :=> 
                     (TGen 0 `fn` TGen 1)))
roundMfun
 = "round" :>: (Forall [Star, Star]
                 ([isIn1 cRealFrac (TGen 0),
                   isIn1 cIntegral (TGen 1)] :=> 
                  (TGen 0 `fn` TGen 1)))
ceilingMfun
 = "ceiling" :>: (Forall [Star, Star]
                   ([isIn1 cRealFrac (TGen 0),
                     isIn1 cIntegral (TGen 1)] :=> 
                    (TGen 0 `fn` TGen 1)))
floorMfun
 = "floor" :>: (Forall [Star, Star]
                 ([isIn1 cRealFrac (TGen 0),
                   isIn1 cIntegral (TGen 1)] :=> 
                  (TGen 0 `fn` TGen 1)))

instsRealFrac
 = [mkInst []
     ([] :=> 
      isIn1 cRealFrac tFloat),
    mkInst []
     ([] :=> 
      isIn1 cRealFrac tDouble),
    mkInst [Star]
     ([isIn1 cIntegral (TGen 0)] :=> 
      isIn1 cRealFrac (TAp tRatio (TGen 0)))]

cRealFloat = Class { name = "RealFloat",
                     csig = asig,
                     super = [isIn1 cRealFrac ta,
                              isIn1 cFloating ta],
                     insts = instsRealFloat }

floatRadixMfun
 = "floatRadix" :>: (Forall [Star]
                      ([isIn1 cRealFloat (TGen 0)] :=> 
                       (TGen 0 `fn` tInteger)))
floatDigitsMfun
 = "floatDigits" :>: (Forall [Star]
                       ([isIn1 cRealFloat (TGen 0)] :=> 
                        (TGen 0 `fn` tInt)))
floatRangeMfun
 = "floatRange" :>: (Forall [Star]
                      ([isIn1 cRealFloat (TGen 0)] :=> 
                       (TGen 0 `fn` TAp (TAp tTuple2 tInt) tInt)))
decodeFloatMfun
 = "decodeFloat" :>: (Forall [Star]
                       ([isIn1 cRealFloat (TGen 0)] :=> 
                        (TGen 0 `fn` TAp (TAp tTuple2 tInteger) tInt)))
encodeFloatMfun
 = "encodeFloat" :>: (Forall [Star]
                       ([isIn1 cRealFloat (TGen 0)] :=> 
                        (tInteger `fn` tInt `fn` TGen 0)))
exponentMfun
 = "exponent" :>: (Forall [Star]
                    ([isIn1 cRealFloat (TGen 0)] :=> 
                     (TGen 0 `fn` tInt)))
significandMfun
 = "significand" :>: (Forall [Star]
                       ([isIn1 cRealFloat (TGen 0)] :=> 
                        (TGen 0 `fn` TGen 0)))
scaleFloatMfun
 = "scaleFloat" :>: (Forall [Star]
                      ([isIn1 cRealFloat (TGen 0)] :=> 
                       (tInt `fn` TGen 0 `fn` TGen 0)))
isNaNMfun
 = "isNaN" :>: (Forall [Star]
                 ([isIn1 cRealFloat (TGen 0)] :=> 
                  (TGen 0 `fn` tBool)))
isInfiniteMfun
 = "isInfinite" :>: (Forall [Star]
                      ([isIn1 cRealFloat (TGen 0)] :=> 
                       (TGen 0 `fn` tBool)))
isDenormalizedMfun
 = "isDenormalized" :>: (Forall [Star]
                          ([isIn1 cRealFloat (TGen 0)] :=> 
                           (TGen 0 `fn` tBool)))
isNegativeZeroMfun
 = "isNegativeZero" :>: (Forall [Star]
                          ([isIn1 cRealFloat (TGen 0)] :=> 
                           (TGen 0 `fn` tBool)))
isIEEEMfun
 = "isIEEE" :>: (Forall [Star]
                  ([isIn1 cRealFloat (TGen 0)] :=> 
                   (TGen 0 `fn` tBool)))
atan2Mfun
 = "atan2" :>: (Forall [Star]
                 ([isIn1 cRealFloat (TGen 0)] :=> 
                  (TGen 0 `fn` TGen 0 `fn` TGen 0)))

instsRealFloat
 = [mkInst []
     ([] :=> 
      isIn1 cRealFloat tFloat),
    mkInst []
     ([] :=> 
      isIn1 cRealFloat tDouble)]

cIx = Class { name = "Ix",
              csig = asig,
              super = [isIn1 cOrd ta],
              insts = instsIx }

rangeMfun
 = "range" :>: (Forall [Star]
                 ([isIn1 cIx (TGen 0)] :=> 
                  (TAp (TAp tTuple2 (TGen 0)) (TGen 0) `fn` TAp tList (TGen 0))))
indexMfun
 = "index" :>: (Forall [Star]
                 ([isIn1 cIx (TGen 0)] :=> 
                  (TAp (TAp tTuple2 (TGen 0)) (TGen 0) `fn` TGen 0 `fn` tInt)))
inRangeMfun
 = "inRange" :>: (Forall [Star]
                   ([isIn1 cIx (TGen 0)] :=> 
                    (TAp (TAp tTuple2 (TGen 0)) (TGen 0) `fn` TGen 0 `fn` tBool)))
rangeSizeMfun
 = "rangeSize" :>: (Forall [Star]
                     ([isIn1 cIx (TGen 0)] :=> 
                      (TAp (TAp tTuple2 (TGen 0)) (TGen 0) `fn` tInt)))

instsIx
 = [mkInst []
     ([] :=> 
      isIn1 cIx tUnit),
    mkInst []
     ([] :=> 
      isIn1 cIx tChar),
    mkInst []
     ([] :=> 
      isIn1 cIx tInt),
    mkInst []
     ([] :=> 
      isIn1 cIx tInteger),
    mkInst []
     ([] :=> 
      isIn1 cIx tBool),
    mkInst []
     ([] :=> 
      isIn1 cIx tOrdering),
    mkInst [Star, Star, Star, Star, Star]
     ([isIn1 cIx (TGen 0),
       isIn1 cIx (TGen 1),
       isIn1 cIx (TGen 2),
       isIn1 cIx (TGen 3),
       isIn1 cIx (TGen 4)] :=> 
      isIn1 cIx (TAp (TAp (TAp (TAp (TAp tTuple5 (TGen 0)) (TGen 1)) (TGen 2)) (TGen 3)) (TGen 4))),
    mkInst [Star, Star, Star, Star]
     ([isIn1 cIx (TGen 0),
       isIn1 cIx (TGen 1),
       isIn1 cIx (TGen 2),
       isIn1 cIx (TGen 3)] :=> 
      isIn1 cIx (TAp (TAp (TAp (TAp tTuple4 (TGen 0)) (TGen 1)) (TGen 2)) (TGen 3))),
    mkInst [Star, Star, Star]
     ([isIn1 cIx (TGen 0),
       isIn1 cIx (TGen 1),
       isIn1 cIx (TGen 2)] :=> 
      isIn1 cIx (TAp (TAp (TAp tTuple3 (TGen 0)) (TGen 1)) (TGen 2))),
    mkInst [Star, Star]
     ([isIn1 cIx (TGen 0),
       isIn1 cIx (TGen 1)] :=> 
      isIn1 cIx (TAp (TAp tTuple2 (TGen 0)) (TGen 1)))]

cEnum = Class { name = "Enum",
                csig = asig,
                super = [],
                insts = instsEnum }

succMfun
 = "succ" :>: (Forall [Star]
                ([isIn1 cEnum (TGen 0)] :=> 
                 (TGen 0 `fn` TGen 0)))
predMfun
 = "pred" :>: (Forall [Star]
                ([isIn1 cEnum (TGen 0)] :=> 
                 (TGen 0 `fn` TGen 0)))
toEnumMfun
 = "toEnum" :>: (Forall [Star]
                  ([isIn1 cEnum (TGen 0)] :=> 
                   (tInt `fn` TGen 0)))
fromEnumMfun
 = "fromEnum" :>: (Forall [Star]
                    ([isIn1 cEnum (TGen 0)] :=> 
                     (TGen 0 `fn` tInt)))
enumFromMfun
 = "enumFrom" :>: (Forall [Star]
                    ([isIn1 cEnum (TGen 0)] :=> 
                     (TGen 0 `fn` TAp tList (TGen 0))))
enumFromThenMfun
 = "enumFromThen" :>: (Forall [Star]
                        ([isIn1 cEnum (TGen 0)] :=> 
                         (TGen 0 `fn` TGen 0 `fn` TAp tList (TGen 0))))
enumFromToMfun
 = "enumFromTo" :>: (Forall [Star]
                      ([isIn1 cEnum (TGen 0)] :=> 
                       (TGen 0 `fn` TGen 0 `fn` TAp tList (TGen 0))))
enumFromThenToMfun
 = "enumFromThenTo" :>: (Forall [Star]
                          ([isIn1 cEnum (TGen 0)] :=> 
                           (TGen 0 `fn` TGen 0 `fn` TGen 0 `fn` TAp tList (TGen 0))))

instsEnum
 = [mkInst []
     ([] :=> 
      isIn1 cEnum tUnit),
    mkInst []
     ([] :=> 
      isIn1 cEnum tChar),
    mkInst []
     ([] :=> 
      isIn1 cEnum tInt),
    mkInst []
     ([] :=> 
      isIn1 cEnum tInteger),
    mkInst []
     ([] :=> 
      isIn1 cEnum tFloat),
    mkInst []
     ([] :=> 
      isIn1 cEnum tDouble),
    mkInst [Star]
     ([isIn1 cIntegral (TGen 0)] :=> 
      isIn1 cEnum (TAp tRatio (TGen 0))),
    mkInst []
     ([] :=> 
      isIn1 cEnum tBool),
    mkInst []
     ([] :=> 
      isIn1 cEnum tOrdering)]

cRead = Class { name = "Read",
                csig = asig,
                super = [],
                insts = instsRead }

readsPrecMfun
 = "readsPrec" :>: (Forall [Star]
                     ([isIn1 cRead (TGen 0)] :=> 
                      (tInt `fn` tReadS (TGen 0))))
readListMfun
 = "readList" :>: (Forall [Star]
                    ([isIn1 cRead (TGen 0)] :=> 
                     tReadS (TAp tList (TGen 0))))

instsRead
 = [mkInst []
     ([] :=> 
      isIn1 cRead tUnit),
    mkInst []
     ([] :=> 
      isIn1 cRead tChar),
    mkInst [Star]
     ([isIn1 cRead (TGen 0)] :=> 
      isIn1 cRead (TAp tList (TGen 0))),
    mkInst []
     ([] :=> 
      isIn1 cRead tInt),
    mkInst []
     ([] :=> 
      isIn1 cRead tInteger),
    mkInst []
     ([] :=> 
      isIn1 cRead tFloat),
    mkInst []
     ([] :=> 
      isIn1 cRead tDouble),
    mkInst [Star]
     ([isIn1 cRead (TGen 0),
       isIn1 cIntegral (TGen 0)] :=> 
      isIn1 cRead (TAp tRatio (TGen 0))),
    mkInst []
     ([] :=> 
      isIn1 cRead tBool),
    mkInst [Star]
     ([isIn1 cRead (TGen 0)] :=> 
      isIn1 cRead (TAp tMaybe (TGen 0))),
    mkInst [Star, Star]
     ([isIn1 cRead (TGen 0),
       isIn1 cRead (TGen 1)] :=> 
      isIn1 cRead (TAp (TAp tEither (TGen 0)) (TGen 1))),
    mkInst []
     ([] :=> 
      isIn1 cRead tOrdering),
    mkInst [Star, Star, Star, Star, Star]
     ([isIn1 cRead (TGen 0),
       isIn1 cRead (TGen 1),
       isIn1 cRead (TGen 2),
       isIn1 cRead (TGen 3),
       isIn1 cRead (TGen 4)] :=> 
      isIn1 cRead (TAp (TAp (TAp (TAp (TAp tTuple5 (TGen 0)) (TGen 1)) (TGen 2)) (TGen 3)) (TGen 4))),
    mkInst [Star, Star, Star, Star]
     ([isIn1 cRead (TGen 0),
       isIn1 cRead (TGen 1),
       isIn1 cRead (TGen 2),
       isIn1 cRead (TGen 3)] :=> 
      isIn1 cRead (TAp (TAp (TAp (TAp tTuple4 (TGen 0)) (TGen 1)) (TGen 2)) (TGen 3))),
    mkInst [Star, Star, Star]
     ([isIn1 cRead (TGen 0),
       isIn1 cRead (TGen 1),
       isIn1 cRead (TGen 2)] :=> 
      isIn1 cRead (TAp (TAp (TAp tTuple3 (TGen 0)) (TGen 1)) (TGen 2))),
    mkInst [Star, Star]
     ([isIn1 cRead (TGen 0),
       isIn1 cRead (TGen 1)] :=> 
      isIn1 cRead (TAp (TAp tTuple2 (TGen 0)) (TGen 1)))]

cShow = Class { name = "Show",
                csig = asig,
                super = [],
                insts = instsShow }

showMfun
 = "show" :>: (Forall [Star]
                ([isIn1 cShow (TGen 0)] :=> 
                 (TGen 0 `fn` tString)))
showsPrecMfun
 = "showsPrec" :>: (Forall [Star]
                     ([isIn1 cShow (TGen 0)] :=> 
                      (tInt `fn` TGen 0 `fn` tShowS)))
showListMfun
 = "showList" :>: (Forall [Star]
                    ([isIn1 cShow (TGen 0)] :=> 
                     (TAp tList (TGen 0) `fn` tShowS)))

instsShow
 = [mkInst []
     ([] :=> 
      isIn1 cShow tUnit),
    mkInst []
     ([] :=> 
      isIn1 cShow tChar),
    mkInst [Star]
     ([isIn1 cShow (TGen 0)] :=> 
      isIn1 cShow (TAp tList (TGen 0))),
    mkInst []
     ([] :=> 
      isIn1 cShow tInt),
    mkInst []
     ([] :=> 
      isIn1 cShow tInteger),
    mkInst []
     ([] :=> 
      isIn1 cShow tFloat),
    mkInst []
     ([] :=> 
      isIn1 cShow tDouble),
    mkInst [Star]
     ([isIn1 cIntegral (TGen 0)] :=> 
      isIn1 cShow (TAp tRatio (TGen 0))),
    mkInst [Star]
     ([] :=> 
      isIn1 cShow (TAp tIO (TGen 0))),
    mkInst []
     ([] :=> 
      isIn1 cShow tIOError),
    mkInst []
     ([] :=> 
      isIn1 cShow tBool),
    mkInst [Star]
     ([isIn1 cShow (TGen 0)] :=> 
      isIn1 cShow (TAp tMaybe (TGen 0))),
    mkInst [Star, Star]
     ([isIn1 cShow (TGen 0),
       isIn1 cShow (TGen 1)] :=> 
      isIn1 cShow (TAp (TAp tEither (TGen 0)) (TGen 1))),
    mkInst []
     ([] :=> 
      isIn1 cShow tOrdering),
    mkInst [Star, Star, Star, Star, Star]
     ([isIn1 cShow (TGen 0),
       isIn1 cShow (TGen 1),
       isIn1 cShow (TGen 2),
       isIn1 cShow (TGen 3),
       isIn1 cShow (TGen 4)] :=> 
      isIn1 cShow (TAp (TAp (TAp (TAp (TAp tTuple5 (TGen 0)) (TGen 1)) (TGen 2)) (TGen 3)) (TGen 4))),
    mkInst [Star, Star, Star, Star]
     ([isIn1 cShow (TGen 0),
       isIn1 cShow (TGen 1),
       isIn1 cShow (TGen 2),
       isIn1 cShow (TGen 3)] :=> 
      isIn1 cShow (TAp (TAp (TAp (TAp tTuple4 (TGen 0)) (TGen 1)) (TGen 2)) (TGen 3))),
    mkInst [Star, Star, Star]
     ([isIn1 cShow (TGen 0),
       isIn1 cShow (TGen 1),
       isIn1 cShow (TGen 2)] :=> 
      isIn1 cShow (TAp (TAp (TAp tTuple3 (TGen 0)) (TGen 1)) (TGen 2))),
    mkInst [Star, Star]
     ([isIn1 cShow (TGen 0),
       isIn1 cShow (TGen 1)] :=> 
      isIn1 cShow (TAp (TAp tTuple2 (TGen 0)) (TGen 1)))]

cFunctor = Class { name = "Functor",
                   csig = msig,
                   super = [],
                   insts = instsFunctor }

fmapMfun
 = "fmap" :>: (Forall [Kfun Star Star, Star, Star]
                ([isIn1 cFunctor (TGen 0)] :=> 
                 ((TGen 1 `fn` TGen 2) `fn` TAp (TGen 0) (TGen 1) `fn` TAp (TGen 0) (TGen 2))))

instsFunctor
 = [mkInst []
     ([] :=> 
      isIn1 cFunctor tMaybe),
    mkInst []
     ([] :=> 
      isIn1 cFunctor tList),
    mkInst []
     ([] :=> 
      isIn1 cFunctor tIO)]

cMonad = Class { name = "Monad",
                 csig = msig,
                 super = [],
                 insts = instsMonad }

returnMfun
 = "return" :>: (Forall [Kfun Star Star, Star]
                  ([isIn1 cMonad (TGen 0)] :=> 
                   (TGen 1 `fn` TAp (TGen 0) (TGen 1))))
mbindMfun
 = ">>=" :>: (Forall [Kfun Star Star, Star, Star]
               ([isIn1 cMonad (TGen 0)] :=> 
                (TAp (TGen 0) (TGen 1) `fn` (TGen 1 `fn` TAp (TGen 0) (TGen 2)) `fn` TAp (TGen 0) (TGen 2))))
mthenMfun
 = ">>" :>: (Forall [Kfun Star Star, Star, Star]
              ([isIn1 cMonad (TGen 0)] :=> 
               (TAp (TGen 0) (TGen 1) `fn` TAp (TGen 0) (TGen 2) `fn` TAp (TGen 0) (TGen 2))))
failMfun
 = "fail" :>: (Forall [Kfun Star Star, Star]
                ([isIn1 cMonad (TGen 0)] :=> 
                 (tString `fn` TAp (TGen 0) (TGen 1))))

instsMonad
 = [mkInst []
     ([] :=> 
      isIn1 cMonad tMaybe),
    mkInst []
     ([] :=> 
      isIn1 cMonad tList),
    mkInst []
     ([] :=> 
      isIn1 cMonad tIO),
    mkInst []
     ([] :=> 
      isIn1 cMonad tTI)]

cHasKind = Class { name = "HasKind",
                   csig = asig,
                   super = [],
                   insts = instsHasKind }

kindMfun
 = "kind" :>: (Forall [Star]
                ([isIn1 cHasKind (TGen 0)] :=> 
                 (TGen 0 `fn` tKind)))

instsHasKind
 = [mkInst []
     ([] :=> 
      isIn1 cHasKind tTyvar),
    mkInst []
     ([] :=> 
      isIn1 cHasKind tTycon),
    mkInst []
     ([] :=> 
      isIn1 cHasKind tType)]

cTypes = Class { name = "Types",
                 csig = asig,
                 super = [],
                 insts = instsTypes }

applyMfun
 = "apply" :>: (Forall [Star]
                 ([isIn1 cTypes (TGen 0)] :=> 
                  (tSubst `fn` TGen 0 `fn` TGen 0)))
tvMfun
 = "tv" :>: (Forall [Star]
              ([isIn1 cTypes (TGen 0)] :=> 
               (TGen 0 `fn` TAp tList tTyvar)))

instsTypes
 = [mkInst []
     ([] :=> 
      isIn1 cTypes tType),
    mkInst [Star]
     ([isIn1 cTypes (TGen 0)] :=> 
      isIn1 cTypes (TAp tList (TGen 0))),
    mkInst [Star]
     ([isIn1 cTypes (TGen 0)] :=> 
      isIn1 cTypes (TAp tQual (TGen 0))),
    mkInst []
     ([] :=> 
      isIn1 cTypes tPred),
    mkInst []
     ([] :=> 
      isIn1 cTypes tScheme),
    mkInst []
     ([] :=> 
      isIn1 cTypes tAssump)]

cInstantiate = Class { name = "Instantiate",
                       csig = asig,
                       super = [],
                       insts = instsInstantiate }

instMfun
 = "inst" :>: (Forall [Star]
                ([isIn1 cInstantiate (TGen 0)] :=> 
                 (TAp tList tType `fn` TGen 0 `fn` TGen 0)))

instsInstantiate
 = [mkInst []
     ([] :=> 
      isIn1 cInstantiate tType),
    mkInst [Star]
     ([isIn1 cInstantiate (TGen 0)] :=> 
      isIn1 cInstantiate (TAp tList (TGen 0))),
    mkInst [Star]
     ([isIn1 cInstantiate (TGen 0)] :=> 
      isIn1 cInstantiate (TAp tQual (TGen 0))),
    mkInst []
     ([] :=> 
      isIn1 cInstantiate tPred)]


-----------------------------------------------------------------------------

defaults    :: [Type]
defaults     = [tInteger, tDouble]

stdClasses  :: [Class]
stdClasses   = [cEq, cOrd, cBounded, cEnum, cIx,
                cShow, cRead,
                cFunctor, cMonad] ++
               numClasses

numClasses  :: [Class]
numClasses   = [cNum, cReal,
                cIntegral, cFractional, cFloating,
                cRealFrac, cRealFloat]

-----------------------------------------------------------------------------
