-----------------------------------------------------------------------------
-- Unify:	Unification
-- 
-- Part of `Typing Haskell in Haskell', version of September 1, 1999
-- Copyright (c) Mark P Jones and the Oregon Graduate Institute
-- of Science and Technology, 1999
-- 
-- This program is distributed as Free Software under the terms
-- in the file "License" that is included in the distribution
-- of this software, copies of which may be obtained from:
--             http://www.cse.ogi.edu/~mpj/thih/
-- 
-----------------------------------------------------------------------------

module Unify where
import Kind
import Type
import Subst

class Unify t where
  mgu :: t -> t -> Maybe Subst

varBind :: Tyvar -> Type -> Maybe Subst

instance Unify Type where
  mgu (TAp l r) (TAp l' r') = do s1 <- mgu l l'
                                 s2 <- mgu (apply s1 r)
                                           (apply s1 r')
                                 Just (s2 @@ s1)
  mgu (TVar u) t        = varBind u t
  mgu t (TVar u)        = varBind u t
  mgu (TCon tc1) (TCon tc2)
             | tc1==tc2 = Just nullSubst
  mgu t1 t2             = Nothing

instance (Types t, Unify t) => Unify [t] where
  mgu (x:xs) (y:ys) = do s1 <- mgu x y
                         s2 <- mgu (apply s1 xs)
                                   (apply s1 ys)
                         return (s2 @@ s1)
  mgu []     []     = return nullSubst
  mgu _      _      = Nothing

varBind u t | t == TVar u      = Just nullSubst
            | u `elem` tv t    = Nothing
            | kind u == kind t = Just (u +-> t)
            | otherwise        = Nothing

class Match t where
  match :: t -> t -> Maybe Subst

instance Match Type where
  match (TAp l r) (TAp l' r') = do sl <- match l l'
                                   sr <- match r r'
                                   merge sl sr
  match (TVar u) t
     | kind u == kind t = Just (u +-> t)
  match (TCon tc1) (TCon tc2)
     | tc1==tc2         = Just nullSubst
  match t1 t2           = Nothing

instance Match t => Match [t] where
  match ts ts' = mergeAll (zipWith match ts ts')
   where mergeAll = foldr cons (Just nullSubst)
         cons (Just s) (Just s') = merge s s'
         cons _        _         = Nothing

-----------------------------------------------------------------------------
